"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Tuple, List, Dict

import numpy as np
import pytest
import rasterio as rio
from rasterio import Affine
from rasterio.windows import Window

from homonim import utils
from homonim.matched_pair import MatchedPairReader


@pytest.mark.parametrize(['file', 'bands', 'exp_bands', 'exp_band_names', 'exp_wavelengths'], [
    ('file_rgba', None, [1, 2, 3], ['1', '2', '3'], [.650, .560, .480]),
    ('file_rgba', (1, 2, 3), [1, 2, 3], ['1', '2', '3'], [.650, .560, .480]),
    ('file_rgba', (1, 2, 3), [1, 2, 3], ['1', '2', '3'], [.650, .560, .480]),
    ('file_masked', None, [1], ['1'], [float('nan')]),
    ('src_file_100cm_float', [1], [1], ['1'], [float('nan')]),
    ('s2_ref_file', None, [1, 2, 3], ['B4', 'B3', 'B2'], [0.6645, 0.56, 0.4966]),
    ('s2_ref_file', (3, 2, 1), [3, 2, 1], ['B2', 'B3', 'B4'], [0.4966, 0.56, 0.6645]),
    (
        'landsat_ref_file', None, list(range(1, 8)) + [9], [f'SR_B{i}' for i in range(1, 8)] + ['ST_B10'],
        [0.443, 0.482, 0.562, 0.655, 0.865, 1.609, 2.201, 10.895]
    ),
    ('landsat_ref_file', [7, 8], [7, 8], ['SR_B7', 'SR_QA_AEROSOL'], [2.201, float('nan')]),
    ('file_bgr', None, [1, 2, 3], ['1', '2', '3'], [ .480, .560, .650]),
    ('file_bgr', (3, 2, 1), [3, 2, 1], ['3', '2', '1'], [.650, .560, .480]),
])  # yapf: disable
def test_get_band_info(
    file: str, bands: Tuple, exp_bands: List, exp_band_names: List, exp_wavelengths: List,
    request: pytest.FixtureRequest
):
    """ Test _get_band_info returns expected values with different image files. """
    file: Path = request.getfixturevalue(file)
    with rio.open(file, 'r') as im:
        bands, band_names, wavelengths = MatchedPairReader._get_band_info(im, bands=bands)
        assert all(bands == np.array(exp_bands))
        assert all(band_names == np.array(exp_band_names))
        assert all(utils.nan_equals(wavelengths, np.array(exp_wavelengths)))


def test_get_band_info_alpha_band_error(file_rgba):
    """ Test an error is raised in _get_band_info when user bands contain alpha bands. """
    with rio.open(file_rgba, 'r') as im:
        with pytest.raises(ValueError) as ex:
            _, _, _ = MatchedPairReader._get_band_info(im, bands=(4, 1, 2))
        assert 'bands contain alpha band(s)' in str(ex)


def test_get_band_info_invalid_band_error(file_rgba):
    """ Test an error is raised in _get_band_info when user bands contain invalid bands. """
    with rio.open(file_rgba, 'r') as im:
        with pytest.raises(ValueError) as ex:
            _, _, _ = MatchedPairReader._get_band_info(im, bands=(4, 1, 0, 2, 5))
        assert 'bands contain invalid band(s)' in str(ex)


@pytest.mark.parametrize(
    ['src_file', 'ref_file', 'src_bands', 'ref_bands', 'exp_src_bands', 'exp_ref_bands', 'force'], [
        ('src_file_100cm_float', 'ref_file_100cm_float', None, None, [1], [1], False),
        ('src_file_100cm_float', 'ref_file_100cm_float', (1,), (1,), [1], [1], False),
        ('src_file_100cm_float', 'file_rgba', None, [1], [1], [1], False),
        ('file_rgba', 'file_rgba', None, None, [1, 2, 3], [1, 2, 3], False),
        ('file_rgba', 'file_rgba', [3, 2], None, [3, 2], [3, 2], False),
        ('file_rgba', 'file_rgba', [3, 2], [3, 1, 2], [3, 2], [3, 2], False),
        ('s2_ref_file', 's2_ref_file', None, None, [1, 2, 3], [1, 2, 3], False),
        ('s2_ref_file', 's2_ref_file', [3, 2], None, [3, 2], [3, 2], False),
        ('s2_ref_file', 'landsat_ref_file', None, None, [1, 2, 3], [4, 3, 2], False),
        ('s2_ref_file', 'landsat_ref_file', [1, 2], list(range(1, 9)), [1, 2], [4, 3], False),
        # ref with mix of spectral and non-alpha bands, len(src) == len(ref)
        ('s2_ref_file', 'landsat_ref_file', [1, 2], [3, 8], [1, 2], [8, 3], False),
        # src & ref with mix of spectral and non-alpha bands, len(src) == len(ref)
        ('landsat_src_file', 'landsat_ref_file', [7, 8, 9, 10], [7, 9, 10, 11], [7, 8, 9, 10], [7, 10, 9, 11], False),
        # src & ref with mix of spectral and non-alpha bands, len(src) < len(ref) & force
        ('landsat_src_file', 'landsat_ref_file', [7, 8, 9, 10], list(range(1, 12)), [7, 8, 9, 10], [1, 2, 3, 4], True),
        # len(src) > len(ref) & force
        ('landsat_src_file', 's2_ref_file', None, None, [1, 2, 3], [1, 2, 3], True),
    ]
)  # yapf: disable
def test_match(
    src_file: str, ref_file: str, src_bands: Tuple, ref_bands: Tuple, exp_src_bands: List, exp_ref_bands: List,
    force: bool, request: pytest.FixtureRequest
):
    """ Test matching of different source and reference files. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with MatchedPairReader(src_file, ref_file, src_bands=src_bands, ref_bands=ref_bands, force=force) as matched_pair:
        assert all(np.array(matched_pair.src_bands) == exp_src_bands)
        assert all(np.array(matched_pair.ref_bands) == exp_ref_bands)


def test_match_fewer_ref_bands_error(s2_ref_file, landsat_ref_file):
    """  Test an error is raised if num src bands > num ref bands. """
    with pytest.raises(ValueError) as ex:
        with MatchedPairReader(landsat_ref_file, s2_ref_file) as matched_pair:
            pass
    assert 'has fewer bands than' in str(ex)


def test_match_wavelength_dist_error(s2_ref_file, landsat_ref_file):
    """  Test an error is raised if src/ref spectral band wavelengths are too far apart. """
    with pytest.raises(ValueError) as ex:
        with MatchedPairReader(s2_ref_file, landsat_ref_file, ref_bands=[1, 3, 5]) as matched_pair:
            pass
    assert 'could not be auto-matched' in str(ex)

@pytest.mark.parametrize(
    ['src_file', 'ref_file', 'src_bands', 'ref_bands','force'], [
        ('src_file_100cm_float', 'file_rgba', None, None, False),
        ('landsat_src_file', 'landsat_ref_file', [7, 8, 9], None, False),
    ]
)  # yapf: disable
def test_match_error(
    src_file: str, ref_file: str, src_bands: Tuple, ref_bands: Tuple, force: bool, request: pytest.FixtureRequest
):
    """  Test an error is raised if src/ref spectral band wavelengths are too far apart. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with pytest.raises(ValueError) as ex:
        with MatchedPairReader(
            src_file, ref_file, src_bands=src_bands, ref_bands=ref_bands, force=force
        ) as matched_pair:
            pass
    assert 'Could not match' in str(ex)



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Tuple

import cv2
import numpy as np
import pytest
from rasterio.enums import Resampling

from homonim.enums import Model
from homonim.kernel_model import SrcSpaceModel, RefSpaceModel
from homonim.raster_array import RasterArray


@pytest.mark.parametrize(
    'model, kernel_shape', [
        (Model.gain, (1, 1)),
        (Model.gain, (3, 3)),
        (Model.gain_blk_offset, (1, 1)),
        (Model.gain_blk_offset, (5, 5)),
        (Model.gain_offset, (5, 5)),
    ]
)  # yapf: disable
def test_ref_basic_fit(
    ra_100cm_float, ra_50cm_float, model: Model, kernel_shape: Tuple[int, int],
):
    """ Test that ref-space models are fitted correctly against known parameters. """

    # mask_partial is only applied in RefSpaceModel.apply(), so we just set it to False here
    kernel_model = RefSpaceModel(model, kernel_shape, mask_partial=False, r2_inpaint_thresh=0.25)
    src_ra = ra_50cm_float
    ref_ra = ra_100cm_float

    param_ra = kernel_model.fit(src_ra, ref_ra.copy())
    assert (param_ra.shape == ref_ra.shape)
    assert (param_ra.transform == ref_ra.transform)
    assert (ref_ra.mask == param_ra.mask).all()  # given that ref_ra.mask == downsample(src_ra.mask)
    assert param_ra.array[0, param_ra.mask] == pytest.approx(1, abs=1.e-2)  # test gain ~1
    assert param_ra.array[1, param_ra.mask] == pytest.approx(0, abs=1.e-2)  # test offset ~0


@pytest.mark.parametrize(
    'model, kernel_shape', [
        (Model.gain, (1, 1)),
        (Model.gain, (3, 3)),
        (Model.gain_blk_offset, (1, 1)),
        (Model.gain_blk_offset, (5, 5)),
        (Model.gain_offset, (5, 5)),
    ]
)  # yapf: disable
def test_src_basic_fit(
    ra_100cm_float, ra_50cm_float, model: Model, kernel_shape: Tuple[int, int],
):
    """ Test that src-space models are fitted correctly against known parameters. """
    kernel_model = SrcSpaceModel(model, kernel_shape, mask_partial=False, r2_inpaint_thresh=0.25)
    src_ra = ra_100cm_float
    ref_ra = ra_50cm_float

    param_ra = kernel_model.fit(src_ra, ref_ra)
    assert (param_ra.shape == src_ra.shape)
    assert (param_ra.transform == src_ra.transform)
    assert (src_ra.mask == param_ra.mask).all()  # given that src_ra.mask == upsample(ref_ra.mask)
    assert param_ra.array[0, param_ra.mask] == pytest.approx(1, abs=1e-2)  # gain ~1
    assert param_ra.array[1, param_ra.mask] == pytest.approx(0, abs=1e-2)  # offset ~0


def test_ref_basic_apply(ra_100cm_float, ra_50cm_float):
    """ Test application of known ref-space parameters. """
    kernel_model = RefSpaceModel(Model.gain_blk_offset, (5, 5), mask_partial=False)
    src_ra = ra_50cm_float

    # create test parameters == 1
    param_ra = ra_100cm_float.copy()
    param_mask = param_ra.mask
    param_ra.array = np.ones((2, *param_ra.shape), dtype=param_ra.dtype)
    param_ra.mask = param_mask

    out_ra = kernel_model.apply(src_ra, param_ra)
    assert (out_ra.transform == src_ra.transform)
    assert (out_ra.shape == src_ra.shape)
    assert (src_ra.mask == out_ra.mask).all()
    assert (out_ra.array[out_ra.mask] == pytest.approx(src_ra.array[out_ra.mask] + 1, abs=1e-2))


def test_src_basic_apply(ra_100cm_float):
    """ Test application of known src-space parameters. """
    kernel_model = SrcSpaceModel(Model.gain_blk_offset, (5, 5), mask_partial=False)
    src_ra = ra_100cm_float

    # create test parameters == 1
    param_ra = ra_100cm_float.copy()
    param_mask = param_ra.mask
    param_ra.array = np.ones((2, *param_ra.shape), dtype=param_ra.dtype)
    param_ra.mask = param_mask

    out_ra = kernel_model.apply(src_ra, param_ra)
    assert (out_ra.transform == src_ra.transform)
    assert (out_ra.shape == src_ra.shape)
    assert (src_ra.mask == out_ra.mask).all()
    assert (out_ra.array[out_ra.mask] == pytest.approx(src_ra.array[out_ra.mask] + 1, abs=1e-2))


@pytest.mark.parametrize(
    'model, find_r2', [
        (Model.gain, True),
        (Model.gain, False),
        (Model.gain_offset, True),
        (Model.gain_offset, False),
        (Model.gain_blk_offset, True),
        (Model.gain_blk_offset, False),
    ]
)  # yapf: disable
def test_ref_find_r2(ra_100cm_float, ra_50cm_float, model: Model, find_r2: bool):
    """ Test ref-space R2 band is added correctly with find_r2==True. """
    kernel_model = RefSpaceModel(model, (5, 5), find_r2=find_r2, r2_inpaint_thresh=None)
    src_ra = ra_50cm_float
    ref_ra = ra_100cm_float
    param_ra = kernel_model.fit(src_ra, ref_ra)
    if find_r2:
        assert (param_ra.count == 3)
        assert np.nanmax(param_ra.array[2]) <= 1
    else:
        assert (param_ra.count == 2)


@pytest.mark.parametrize(
    'model, find_r2', [
        (Model.gain, True),
        (Model.gain, False),
        (Model.gain_offset, True),
        (Model.gain_offset, False),
        (Model.gain_blk_offset, True),
        (Model.gain_blk_offset, False),
    ]
)  # yapf: disable
def find_r2(float_100cm_ra: RasterArray, float_50cm_ra: RasterArray, model: Model, find_r2: bool):
    """ Test src-space R2 band is added correctly with find_r2==True. """
    kernel_model = SrcSpaceModel(model, (5, 5), find_r2=find_r2, r2_inpaint_thresh=None)
    src_ra = float_100cm_ra
    ref_ra = float_50cm_ra
    param_ra = kernel_model.fit(ref_ra, src_ra)
    if find_r2:
        assert (param_ra.count == 3)
        assert np.nanmax(param_ra.array[2]) <= 1
    else:
        assert (param_ra.count == 2)


@pytest.mark.parametrize('kernel_shape', [(5, 5), (5, 7), (9, 9)])
def test_r2_inpainting(ra_50cm_float, kernel_shape: Tuple[int, int]):
    """ Test R2 values and in-painting. """

    # make src and ref the same so we have known parameters == 1
    src_ra = ra_50cm_float
    ref_ra = ra_50cm_float.copy()

    # find indices and masks to set a ref pixel to -100, so that r2 values are low for all kernels covering that pixel
    low_r2_loc = np.floor(np.array(ref_ra.shape) / 2).astype('int')
    low_r2_ul = (low_r2_loc - np.floor((np.array(kernel_shape) / 2))).astype('int')
    low_r2_mask = np.zeros_like(ref_ra.mask).astype('bool')
    low_r2_mask[low_r2_ul[0]:low_r2_ul[0] + kernel_shape[0], low_r2_ul[1]:low_r2_ul[1] + kernel_shape[1]] = True
    ref_ra.array[low_r2_loc[0], low_r2_loc[1]] = -100

    # fit models with and without inpainting
    no_inpaint_kernel_model = RefSpaceModel(
        Model.gain_offset, kernel_shape=kernel_shape, r2_inpaint_thresh=-np.inf, mask_partial=False
    )
    no_inpaint_param_ra = no_inpaint_kernel_model.fit(src_ra, ref_ra.copy())

    inpaint_kernel_model = RefSpaceModel(
        Model.gain_offset, kernel_shape=kernel_shape, r2_inpaint_thresh=0.5, mask_partial=False
    )
    inpaint_param_ra = inpaint_kernel_model.fit(src_ra, ref_ra.copy())

    # test r2 values
    for param_ra in [no_inpaint_param_ra, inpaint_param_ra]:
        assert (param_ra.array[2, ~low_r2_mask & ref_ra.mask] == pytest.approx(1, abs=1.e-3))
        assert (param_ra.array[2, low_r2_mask] < .5).all()

    # test r2 inpainting has improved parameters
    assert (no_inpaint_param_ra.array[1, no_inpaint_param_ra.mask] != pytest.approx(0, abs=1.e-1))
    assert (inpaint_param_ra.array[1, inpaint_param_ra.mask] == pytest.approx(0, abs=1.e-1))
    assert (
        inpaint_param_ra.array[0, inpaint_param_ra.mask].var() <
        no_inpaint_param_ra.array[0, no_inpaint_param_ra.mask].var()
    )


@pytest.mark.parametrize(
    'kernel_shape, mask_partial', [
        ((1, 1), False),
        ((1, 1), True),
        ((3, 3), True),
        ((3, 5), True),
        ((5, 5), True),
    ]
)  # yapf: disable
def test_ref_masking(
    ra_100cm_float, ra_50cm_float, kernel_shape: Tuple[int, int], mask_partial: bool,
):
    """ Test ref-space partial masking. """
    kernel_model = RefSpaceModel(Model.gain_blk_offset, kernel_shape, mask_partial=mask_partial)
    src_ra = ra_50cm_float.copy()

    # create test parameters == 1
    param_ra = ra_100cm_float.copy()
    param_mask = param_ra.mask
    param_ra.array = np.ones((2, *param_ra.shape), dtype=param_ra.dtype)
    param_ra.mask = param_mask

    out_ra = kernel_model.apply(src_ra, param_ra)
    if not mask_partial:
        assert (src_ra.mask == out_ra.mask).all()
    else:
        # test output mask is contained by and smaller than src mask
        assert (src_ra.mask.sum() > out_ra.mask.sum())
        assert src_ra.mask[out_ra.mask].all()

        # find and test against the expected mask
        # this test depends on RasterArray.reproject which is a compromise to allow thorough testing here
        mask_ra = src_ra.mask_ra.reproject(**param_ra.proj_profile, nodata=None, resampling=Resampling.average)
        mask = (mask_ra.array >= 1).astype('uint8', copy=False)  # ref pixels fully covered by src
        mask_ra.array = cv2.erode(mask, np.ones(np.add(kernel_shape, 2)))
        test_mask_ra = mask_ra.reproject(**src_ra.proj_profile, nodata=0, resampling=Resampling.nearest)
        assert (test_mask_ra.array == out_ra.mask).all()


@pytest.mark.parametrize(
    'kernel_shape, mask_partial', [
        ((1, 1), False),
        ((1, 1), True),
        ((3, 3), True),
        ((3, 5), True),
        ((5, 5), True),
    ]
)  # yapf: disable
def test_src_masking(
    ra_100cm_float, ra_50cm_float, kernel_shape: Tuple[int, int], mask_partial: bool,
):
    """ Test src-space partial masking. """
    kernel_model = SrcSpaceModel(Model.gain_blk_offset, kernel_shape, mask_partial=mask_partial)
    src_ra = ra_100cm_float
    ref_ra = ra_50cm_float

    # create test parameters
    param_ra = kernel_model.fit(src_ra, ref_ra)

    if not mask_partial:
        assert (src_ra.mask == param_ra.mask).all()
    else:
        # test output mask is contained by and smaller than src mask
        assert (src_ra.mask.sum() > param_ra.mask.sum())
        assert src_ra.mask[param_ra.mask].all()
        # find and test against the expected mask
        test_mask = cv2.erode(src_ra.mask.astype('uint8'), np.ones(np.array(kernel_shape) + 2))
        assert (test_mask == param_ra.mask).all()


def test_ref_force_proc_crs(ra_100cm_float, ra_50cm_float):
    """ Test fitting models in ref space with low res src. """
    kernel_model = RefSpaceModel(Model.gain_blk_offset, (5, 5), mask_partial=False)
    src_ra = ra_100cm_float
    ref_ra = ra_50cm_float
    param_ra = kernel_model.fit(src_ra, ref_ra.copy())
    out_ra = kernel_model.apply(src_ra, param_ra)
    assert (src_ra.array[src_ra.mask] == pytest.approx(out_ra.array[out_ra.mask], abs=2))


def test_src_force_proc_crs(ra_100cm_float, ra_50cm_float):
    """ Test fitting models in src space with low res ref. """
    kernel_model = SrcSpaceModel(Model.gain_blk_offset, (5, 5), mask_partial=False)
    src_ra = ra_50cm_float
    ref_ra = ra_100cm_float
    param_ra = kernel_model.fit(src_ra, ref_ra)
    out_ra = kernel_model.apply(src_ra, param_ra)
    assert (src_ra.array[src_ra.mask] == pytest.approx(out_ra.array[out_ra.mask], abs=2))


@pytest.mark.parametrize(
    'model, kernel_shape', [
        (Model.gain, (0, 0)),
        (Model.gain_blk_offset, (0, 0)),
        (Model.gain_offset, (4, 5)),
    ]
)  # yapf: disable
def test_kernel_shape_exception(model: Model, kernel_shape: Tuple[int, int]):
    """ Test bad kernel shape / model combination raises an exception. """
    with pytest.raises(ValueError):
        _ = RefSpaceModel(model=model, kernel_shape=kernel_shape)


def test_r2_array_defaults(ra_100cm_float):
    """ Test KernelModel._r2_array() with default (none) parameter values. """
    kernel_model = RefSpaceModel(Model.gain_offset, (5, 5), mask_partial=False, r2_inpaint_thresh=None, find_r2=False)
    src_ra = ra_100cm_float
    ref_ra = ra_100cm_float
    param_ra = kernel_model.fit(src_ra, ref_ra.copy())
    r2_array = kernel_model._r2_array(ref_ra.array, ref_ra.array, param_ra.array)
    assert (r2_array[~np.isnan(r2_array)] == pytest.approx(1, abs=1.e-3))


def test_config():
    """ Test KernelModel is configured correctly.  """
    config = dict(
        r2_inpaint_thresh=0.1, mask_partial=True, downsampling=Resampling.bilinear, upsampling=Resampling.nearest,
    )
    kernel_model = RefSpaceModel(Model.gain, (5, 5), find_r2=True, **config)
    for key in config.keys():
        attr_name = '_' + key
        assert hasattr(kernel_model, attr_name)
        assert getattr(kernel_model, attr_name) == config[key]


def test_config_exception():
    """ Test an Exception is raised if an unknown config kwarg is passed.  """
    with pytest.raises(TypeError):
        kernel_model = RefSpaceModel(Model.gain, (5, 5), find_r2=True, unknown='value')



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import multiprocessing
from pathlib import Path
from typing import List, Dict, Tuple

import numpy as np
import pytest
from click.testing import CliRunner
from pytest import FixtureRequest

from homonim.cli import cli
from homonim.compare import RasterCompare
from homonim.enums import ProcCrs
from tests.conftest import str_contain_no_space


def _test_identical_compare_dict(res_dict: Dict, exp_len: int = 4):
    """ Helper function to run tests on a compare results list, where the compare was between identical images. """
    assert (len(res_dict) == exp_len)
    bands = list(res_dict.keys())
    assert (bands[-1] == 'Mean')
    band_list = list(res_dict.values())[:-1]

    r2 = np.array([res_item['r2'] for res_item in band_list])
    rmse = np.array([res_item['rmse'] for res_item in band_list])
    rrmse = np.array([res_item['rrmse'] for res_item in band_list])
    n = np.array([res_item['n'] for res_item in band_list])
    assert (r2 == pytest.approx(1))
    assert (n == n[0]).all()
    assert (rmse == pytest.approx(0))
    assert (rrmse == pytest.approx(0))
    assert (res_dict['Mean']['r2'] == pytest.approx(1))
    assert (res_dict['Mean']['rmse'] == pytest.approx(0))
    assert (res_dict['Mean']['rrmse'] == pytest.approx(0))


@pytest.mark.parametrize(
    'src_file, ref_file, src_bands, ref_bands, force, exp_bands', [
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', None, None, False, (1, 2, 3)),
        ('file_rgb_100cm_float', 'file_rgb_50cm_float', (3, 2, 1), None, False, (3, 2, 1)),
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', (2, 1), (3, 1, 2), False, (2, 1)),
        ('file_rgb_100cm_float', 'file_rgb_50cm_float', (2, 1), (3, 2, 1), True, (3, 2)),
    ]
)  # yapf: disable
def test_api(
    src_file: str, ref_file: str, src_bands: Tuple[int], ref_bands: Tuple[int], force: bool, exp_bands: Tuple[int],
    tmp_path: Path, request: FixtureRequest
):
    """ Test fusion with the src_bands and ref_bands parameters. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with RasterCompare(src_file, ref_file, src_bands=src_bands, ref_bands=ref_bands, force=force) as raster_compare:
        res_dict = raster_compare.process()
        assert raster_compare.ref_bands == exp_bands
    if not force:
        _test_identical_compare_dict(res_dict, len(exp_bands) + 1)


def test_api__thread(src_file_45cm_float, ref_file_100cm_float):
    """ Test compasison results remain the same with different `threads` configurations. """
    with RasterCompare(src_file_45cm_float, ref_file_100cm_float) as raster_compare:
        res_dict_single = raster_compare.process(threads=1)
        res_dict_mult = raster_compare.process(threads=multiprocessing.cpu_count())
    assert (len(res_dict_single) == 2)
    assert (len(res_dict_mult) == 2)
    assert (res_dict_mult == res_dict_single)


@pytest.mark.parametrize(
    'src_file, ref_file, proc_crs, config', [
        ('src_file_50cm_float', 'ref_file_100cm_float', ProcCrs.ref, dict(downsampling='lanczos')),
        ('src_file_50cm_float', 'ref_file_100cm_float', ProcCrs.src, dict(upsampling='lanczos')),
    ]
)  # yapf:disable
def test_api__resampling(src_file: str, ref_file: str, proc_crs: ProcCrs, config: Dict, request: FixtureRequest):
    """ Test non-default resampling parameters give similar but different results to the defaults. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with RasterCompare(src_file, ref_file, proc_crs=proc_crs) as raster_compare:
        res_dict_def = raster_compare.process()  # default configuration results
        res_dict_lz = raster_compare.process(**config)  # non-default configuration
    assert (len(res_dict_def) == 2)
    assert (len(res_dict_lz) == 2)
    # test non-default r2 is similar but different to default r2
    for band in res_dict_def.keys():
        assert res_dict_def[band]['r2'] != pytest.approx(res_dict_lz[band]['r2'], rel=1e-5)
        assert res_dict_def[band]['r2'] == pytest.approx(res_dict_lz[band]['r2'], rel=1e-1)


@pytest.mark.parametrize(
    'src_file, ref_file', [
        ('src_file_100cm_float', 'ref_file_100cm_float'),
        ('src_file_45cm_float', 'ref_file_100cm_float'),
        ('src_file_100cm_float', 'ref_file_45cm_float'),
    ]
)  # yapf:disable
def test_api__max_block_mem(src_file: str, ref_file: str, request: FixtureRequest):
    """ Test changing the number and shape of blocks (i.e. max_block_mem) gives the same comparison results. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with RasterCompare(src_file, ref_file) as compare:
        stats_dict_band = compare.process(max_block_mem=100)  # compare by band
        stats_dict_block = compare.process(max_block_mem=2e-4)  # compare by small block
    assert (len(stats_dict_band) == 2)
    assert (len(stats_dict_block) == 2)
    # test band-based and block-based results are approx the same
    for band in stats_dict_band.keys():
        for k in stats_dict_band[band].keys():
            assert stats_dict_band[band][k] == pytest.approx(stats_dict_block[band][k], rel=1e-5)


def test_api__proc_crs(
    src_file_45cm_float, ref_file_100cm_float, src_file_100cm_float, ref_file_45cm_float
):
    """
    Test comparison of high res source with low res reference (proc_crs=ref) gives approx same results as comparison of
    low res source with high res reference (proc_crs=src).
    """
    with RasterCompare(src_file_45cm_float, ref_file_100cm_float, proc_crs=ProcCrs.ref) as raster_compare:
        stats_dict_ref = raster_compare.process()  # compare by band
        assert (raster_compare.proc_crs == ProcCrs.ref)
    assert (len(stats_dict_ref) == 2)
    with RasterCompare(src_file_100cm_float, ref_file_45cm_float, proc_crs=ProcCrs.src) as raster_compare:
        stats_dict_src = raster_compare.process()  # compare by band
        assert (raster_compare.proc_crs == ProcCrs.src)
    assert (len(stats_dict_src) == 2)
    # test ProcCrs.ref and ProcCrs.src results are approx the same
    for band in stats_dict_ref.keys():
        for k in stats_dict_ref[band].keys():
            assert stats_dict_ref[band][k] == pytest.approx(stats_dict_src[band][k], rel=1e-3)


def test_cli(runner: CliRunner, file_rgb_50cm_float, file_rgb_100cm_float):
    """ Test compare CLI with known outputs. """
    ref_file = file_rgb_100cm_float
    src_file = file_rgb_50cm_float

    cli_str = f'compare {src_file} {ref_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    res_str = """Ref. band 1 1.000  0.000  0.000   144
Ref. band 2 1.000  0.000  0.000   144
Ref. band 3 1.000  0.000  0.000   144
Mean   1.000  0.000  0.000   144"""
    assert (str_contain_no_space(res_str, result.output))


def test_cli__output_file(tmp_path: Path, runner: CliRunner, file_rgb_50cm_float, file_rgb_100cm_float):
    """ Test compare CLI generated json file. """
    ref_file = file_rgb_100cm_float
    src_file = file_rgb_50cm_float

    output_file = tmp_path.joinpath('compare.json')
    cli_str = f'compare {src_file} {ref_file} --output {output_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (output_file.exists())

    with open(output_file) as f:
        stats_dict = json.load(f)

    src_file = str(src_file)
    assert (src_file in stats_dict)
    _test_identical_compare_dict(stats_dict[src_file])


def test_cli__mult_inputs(tmp_path: Path, runner: CliRunner, file_rgb_50cm_float, file_rgb_100cm_float):
    """ Test compare CLI with multiple src files. """
    ref_file = file_rgb_100cm_float
    src_file = file_rgb_50cm_float

    output_file = tmp_path.joinpath('compare.json')
    cli_str = f'compare {src_file} {src_file} {ref_file} --output {output_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (output_file.exists())

    with open(output_file) as f:
        stats_dict = json.load(f)

    src_file = str(src_file)
    assert (src_file in stats_dict)


def test_cli__adv_options(tmp_path: Path, runner: CliRunner, src_file_50cm_float, ref_file_100cm_float):
    """ Test that the combined advanced CLI options affect comparison results as expected. """
    ref_file = ref_file_100cm_float
    src_file = src_file_50cm_float

    # run a comparison with default advanced options, and with specified advanced options, then compare results
    out_file_def = tmp_path.joinpath('compare_defaults.json')
    out_file_adv = tmp_path.joinpath('compare_adv.json')
    cli_str_def = f'compare {src_file} {ref_file} --output {out_file_def}'
    cli_str_adv = f"""compare --threads 1 --max-block-mem 1e-3 --downsampling bilinear -pc ref {src_file} {ref_file}  
    --output {out_file_adv}"""
    stats_list = []
    for cli_str, out_file in zip([cli_str_def, cli_str_adv], [out_file_def, out_file_adv]):
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        assert (out_file.exists())
        with open(out_file) as f:
            stats_dict = json.load(f)
            assert (str(src_file) in stats_dict)
            stats_list.append(stats_dict)
    b1_dict_def = stats_list[0][str(src_file)]['Mean']
    b1_dict_adv = stats_list[1][str(src_file)]['Mean']
    # test that r2 with default options, and r2 with advanced options, are different, but not too different
    assert b1_dict_def['r2'] != pytest.approx(b1_dict_adv['r2'], 1e-5)
    assert b1_dict_def['r2'] == pytest.approx(b1_dict_adv['r2'], 1e-1)


@pytest.mark.parametrize(
    'src_bands, ref_bands, force, exp_bands', [
        ((3, 2, 1), None, False, (3, 2, 1)),
        ((2, 1), (3, 1, 2), False, (2, 1)),
        ((2, 1), (3, 2, 1), True, (3, 2)),
    ]
)  # yapf: disable
def test_cli_src_ref_bands(
    src_bands: Tuple[int], ref_bands: Tuple[int], force: bool, exp_bands: Tuple[int], file_rgb_50cm_float,
    file_rgb_100cm_float, tmp_path: Path, runner: CliRunner,
):
    """ Test compare with --src_band, --ref_band and --force-match parameters. """
    src_file = file_rgb_50cm_float
    ref_file = file_rgb_100cm_float
    out_file = tmp_path.joinpath('results.json')
    cli_str = f'compare {src_file} {ref_file} -op {out_file}'
    if src_bands:
        cli_str += ''.join([' -sb ' + str(bi) for bi in src_bands])
    if ref_bands:
        cli_str += ''.join([' -rb ' + str(bi) for bi in ref_bands])
    if force:
        cli_str += ' -f'

    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (out_file.exists())
    with open(out_file) as f:
        stats_dict = json.load(f)
    assert str(src_file) in stats_dict
    stats_dict = stats_dict[str(src_file)]
    exp_band_names = [f'Ref. band {bi}' for bi in exp_bands] + ['Mean']
    assert list(stats_dict.keys()) == exp_band_names
    if not force:
        _test_identical_compare_dict(stats_dict, len(exp_bands) + 1)




"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import pathlib
from multiprocessing import cpu_count
from typing import Tuple, Dict, Union
from contextlib import contextmanager

import numpy as np
import rasterio as rio
from rasterio.enums import ColorInterp, Resampling
from rasterio.vrt import WarpedVRT
from rasterio.windows import Window
from tabulate import TableFormat, Line, DataRow, tabulate

from homonim.enums import Model, ProcCrs
from homonim.errors import ImageFormatError

logger = logging.getLogger(__name__)
tabulate.MIN_PADDING = 0

##
table_format = TableFormat(
    lineabove=Line("", "-", " ", ""),
    linebelowheader=Line("", "-", " ", ""),
    linebetweenrows=None,
    linebelow=Line("", "-", " ", ""),
    headerrow=DataRow("", " ", ""),
    datarow=DataRow("", " ", ""),
    padding=0,
    with_header_hide=["lineabove", "linebelow"]
)  # yapf: disable
""" Tabulate format for comparison and parameter stats. """


def nan_equals(a: Union[np.ndarray, float], b: Union[np.ndarray, float]) -> np.ndarray:
    """ Compare two numpy objects a & b, returning true where elements of both a & b are nan. """
    return (a == b) | (np.isnan(a) & np.isnan(b))


def expand_window_to_grid(win: Window, expand_pixels: Tuple[int, int] = (0, 0)) -> Window:
    """
    Expand rasterio window extents to the nearest whole numbers i.e. for ``expand_pixels`` >= (0, 0), it will return a
    window that contains the original extents.

    Parameters
    ----------
    win: rasterio.windows.Window
        Window to expand.
    expand_pixels: tuple, optional
        Tuple specifying the number of (rows, columns) pixels to expand the window by.

    Returns
    -------
    rasterio.windows.Window
        Expanded window.
    """
    col_off, col_frac = np.divmod(win.col_off - expand_pixels[1], 1)
    row_off, row_frac = np.divmod(win.row_off - expand_pixels[0], 1)
    width = np.ceil(win.width + 2 * expand_pixels[1] + col_frac)
    height = np.ceil(win.height + 2 * expand_pixels[0] + row_frac)
    exp_win = Window(col_off.astype('int'), row_off.astype('int'), width.astype('int'), height.astype('int'))
    return exp_win


def round_window_to_grid(win: Window) -> Window:
    """
    Round window extents to the nearest whole numbers.

    Parameters
    ----------
    win: rasterio.windows.Window
        Window to round.

    Returns
    -------
    rasterio.windows.Window
        Rounded window with integer extents.
    """
    row_range, col_range = win.toranges()
    row_range = np.round(row_range).astype('int')
    col_range = np.round(col_range).astype('int')
    return Window(col_off=col_range[0], row_off=row_range[0], width=np.diff(col_range)[0], height=np.diff(row_range)[0])


def validate_kernel_shape(kernel_shape: Tuple[int, int], model: Model = Model.gain_blk_offset) -> Tuple[int, int]:
    """
    Check a kernel shape (height, width) tuple for validity.  Raises ValueError if ``kernel_shape`` is invalid.

    Parameters
    ----------
    kernel_shape: tuple of int
        Kernel (height, width) in pixels.
    model: Model, optional
        The model type ``kernel_shape`` will be used with.

    Returns
    -------
    tuple of int
        The validated kernel_shape as a numpy array.
    """
    kernel_shape = np.array(kernel_shape).astype(int)
    if not np.all(np.mod(kernel_shape, 2) == 1):
        raise ValueError('`kernel_shape` must be odd in both dimensions.')
    if model == Model.gain_offset:
        if np.product(kernel_shape) < 2:
            raise ValueError('`kernel_shape` area should contain at least 2 elements for the gain-offset model.')
        elif np.product(kernel_shape) < 25:
            logger.warning('A `kernel_shape` of at least 25 elements is recommended for the gain-offset model.')
    if not np.all(kernel_shape >= 1):
        raise ValueError('`kernel_shape` must be a minimum of one in both dimensions.')
    return tuple(kernel_shape)


def overlap_for_kernel(kernel_shape: Tuple[int, int]) -> Tuple[int, int]:
    """
    Return the block overlap for a kernel shape.

    Parameters
    ----------
    kernel_shape: tuple of int
        Kernel (height, width) in pixels.

    Returns
    -------
    tuple of int
        Overlap (height, width) in integer pixels.
    """
    # Block overlap should be at least half the kernel 'shape' to ensure full kernel coverage at block edges, and a
    # minimum of (1, 1) to avoid including extrapolated (rather than interpolated) pixels when up-sampling.
    kernel_shape = np.array(kernel_shape).astype(int)
    return tuple(np.ceil(kernel_shape / 2).astype('int'))


def validate_threads(threads: int) -> int:
    """ Parse number of threads parameter. """
    _cpu_count = cpu_count()
    threads = _cpu_count if threads == 0 else threads
    if threads > _cpu_count:
        raise ValueError(f"'threads' is limited to the number of processors ({_cpu_count})")
    return threads


def create_out_postfix(proc_crs: ProcCrs, model: Model, kernel_shape: Tuple[int, int], driver: str = 'GTiff') -> str:
    """ Create a filename postfix, including extension, for the corrected image file. """
    ext_dict = rio.drivers.raster_driver_extensions()
    ext_idx = list(ext_dict.values()).index(driver)
    ext = list(ext_dict.keys())[ext_idx]
    post_fix = f'_FUSE_c{proc_crs.name.upper()}_m{model.upper()}_k{kernel_shape[0]}_{kernel_shape[1]}.{ext}'
    return post_fix


def create_param_filename(filename: Union[str, pathlib.Path]) -> pathlib.Path:
    """ Create a debug image filename, given the corrected image filename. """
    filename = pathlib.Path(filename)
    return filename.parent.joinpath(f'{filename.stem}_PARAM{filename.suffix}')


def north_up(im: rio.DatasetReader) -> bool:
    """ Return true if im is in a standard North-up orientation. """
    return(
        (np.sign(im.transform.a) == 1) and (np.sign(im.transform.e) == -1) and (im.transform.b == 0) and
        (im.transform.d == 0)
    )


def same_orientation_crs(
    src_im: rio.DatasetReader, ref_im: rio.DatasetReader, proc_crs: ProcCrs = None,
) -> Tuple[rio.DatasetReader, rio.DatasetReader]:  # yapf: disable
    """
    Re-project source and reference (as necessary) so that are they both oriented north-up, and in the same CRS.
    Re-project the proc_crs image (usually lower resolution) into the CRS of the other when the CRSs are not
    the same.
    """
    # Note: without transform etc arguments, WarpedVRT re-projects to north-up
    resampling = Resampling.bilinear
    same_crs = src_im.crs.to_proj4() == ref_im.crs.to_proj4()
    if (not north_up(src_im)) and (same_crs or proc_crs != ProcCrs.src):
        src_im = WarpedVRT(src_im, crs=src_im.crs, resampling=resampling)
    if (not north_up(ref_im)) and (same_crs or proc_crs == ProcCrs.src):
        ref_im = WarpedVRT(ref_im, crs=ref_im.crs, resampling=resampling)
    if (not same_crs) and (proc_crs == ProcCrs.src):
        src_im = WarpedVRT(src_im, crs=ref_im.crs, resampling=resampling)
    if (not same_crs) and (proc_crs != ProcCrs.src):
        ref_im = WarpedVRT(ref_im, crs=src_im.crs, resampling=resampling)
    return src_im, ref_im


@contextmanager
def same_orientation_crs_ctx(
    src_im: rio.DatasetReader, ref_im: rio.DatasetReader, proc_crs: ProcCrs = None,
) -> Tuple[rio.DatasetReader, rio.DatasetReader]:  # yapf: disable
    """
    Context manager to re-project source and reference (as necessary) so that are they both oriented north-up,
    and in the same CRS.
    """
    try:
        src_im, ref_im = same_orientation_crs(src_im, ref_im, proc_crs=proc_crs)
        yield (src_im, ref_im)
    finally:
        src_im.close()
        ref_im.close()


def covers_bounds(im1: rio.DatasetReader, im2: rio.DatasetReader, expand_pixels: Tuple[int, int] = (0, 0)) -> bool:
    """
    Determines if the spatial extents of one image cover another image

    Parameters
    ----------
    im1: rasterio.DatasetReader
        An open rasterio dataset.
    im2: rasterio.DatasetReader
        Another open rasterio dataset.
    expand_pixels: tuple of int, optional
        Expand the im2 bounds by this many pixels.

    Returns
    -------
    bool
        True if im1 covers im2 else False.
    """
    with same_orientation_crs_ctx(im1, im2) as (im1, im2):
        im1_win = im1.window(*im2.bounds)
    if not np.all(np.array(expand_pixels) == 0):
        im1_win = expand_window_to_grid(im1_win, expand_pixels)
    win_ul = np.array((im1_win.row_off, im1_win.col_off))
    win_shape = np.array((im1_win.height, im1_win.width))
    return False if np.any(win_ul < 0) or np.any(win_shape > im1.shape) else True


def get_nonalpha_bands(im: Union[rio.DatasetReader, rio.io.DatasetWriter]) -> Tuple[int, ...]:
    """
    Return a list of non-alpha band indices from a rasterio dataset.

    Parameters
    ----------
    im: rasterio.DatasetReader
        Retrieve band indices from this dataset.

    Returns
    -------
    list of int
        List of 1-based band indices.
    """
    bands = tuple([bi + 1 for bi in range(im.count) if im.colorinterp[bi] != ColorInterp.alpha])
    return bands


def combine_profiles(in_profile: Dict, config_profile: Dict) -> Dict:
    """
    Update an input rasterio profile with a configuration profile.

    Parameters
    ----------
    in_profile: dict
        Input/initial rasterio profile to update.  Driver-specific items are in the root dict.
    config_profile: dict
        Configuration profile.  Driver specific options are contained in a nested dict,
        with the ``creation_options`` key. E.g. see :meth:`homonim.RasterFuse.create_out_profile`.

    Returns
    -------
    dict
        Combined profile.
    """

    if in_profile['driver'].lower() != config_profile['driver'].lower():
        # copy only non-driver specific keys from input profile when the driver is different to the configured val
        copy_keys = ['driver', 'width', 'height', 'count', 'dtype', 'crs', 'transform']
        out_profile = {copy_key: in_profile[copy_key] for copy_key in copy_keys}
    else:
        out_profile = in_profile.copy()  # copy the whole input profile

    def nested_update(self_dict, other_dict):
        """ Update self_dict with a flattened version of other_dict. """
        for other_key, other_value in other_dict.items():
            if isinstance(other_value, dict):
                # flatten the driver specific nested dict into the root dict
                nested_update(self_dict, other_value)
            # elif other_value is not None:
            else:
                self_dict[other_key] = other_value
        return self_dict

    # update out_profile with a flattened config_profile
    return nested_update(out_profile, config_profile)


def validate_param_image(param_filename: Union[str, pathlib.Path]):
    """ Check file is a valid parameter image. """
    param_filename = pathlib.Path(param_filename)
    if not param_filename.exists():
        raise FileNotFoundError(f'{param_filename} does not exist')

    with rio.open(param_filename) as param_im:
        tags = param_im.tags()
        # check band count is a multiple of 3 and that expected metadata tags exist
        if (param_im.count == 0 or divmod(param_im.count, 3)[1] != 0 or
                not {'FUSE_MODEL', 'FUSE_KERNEL_SHAPE', 'FUSE_PROC_CRS', 'FUSE_REF_FILE'} <= set(tags)):
            raise ImageFormatError(f'{param_filename.name} is not a valid parameter image.')

        # check band descriptions end with the expected suffixes
        n_refl_bands = int(param_im.count / 3)
        suffixes = ['gain'] * n_refl_bands + ['offset'] * n_refl_bands + ['r2'] * n_refl_bands
        if not all([desc.lower().endswith(suffix) for suffix, desc in zip(suffixes, param_im.descriptions)]):
            raise ImageFormatError(f'{param_filename.name} is not a valid parameter image.')



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
from pathlib import Path

from setuptools import setup, find_packages

"""
 Build and upload to testpypi:
     conda install -c conda-forge build twine
     python -m build
     python -m twine upload --repository testpypi dist/*

 Install from testpypi:
    python -m pip install --extra-index-url https://test.pypi.org/simple/ homonim

 Install local development version:
    pip install -e .
"""

this_directory = Path(__file__).parent
long_description = (this_directory / 'README.rst').read_text()
sys.path[0:0] = ['homonim']
from version import __version__

setup(
    name='homonim',
    version=__version__,
    description='Correct aerial and satellite imagery to surface reflectance.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Dugal Harris',
    author_email='dugalh@gmail.com',
    url='https://github.com/leftfield-geospatial/homonim',
    license='AGPLv3',
    packages=find_packages(include=['homonim']),
    install_requires=[
        'numpy>=1.19',
        'rasterio>=1.1',
        'opencv-python-headless>=4.5',
        'click>=8',
        'tqdm>=4.6',
        'pyyaml>=5.4',
        'cloup>=0.15',
        'tabulate>=0.8',
    ],
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
    ],
    keywords=[
        'drone imagery', 'aerial imagery', 'satellite imagery', 'surface reflectance', 'correction', 'harmonization',
        'anisotropy', 'brdf', 'atmospheric correction',
    ],
    entry_points={'console_scripts': ['homonim=homonim.cli:cli']},
    project_urls={
        'Documentation': 'https://homonim.readthedocs.io',
        'Source': 'https://github.com/leftfield-geospatial/homonim',
    },
)  # yapf: disable
